/*
This file is part of CASToR.

    CASToR is free software: you can redistribute it and/or modify it under the
    terms of the GNU General Public License as published by the Free Software
    Foundation, either version 3 of the License, or (at your option) any later
    version.

    CASToR is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
    details.

    You should have received a copy of the GNU General Public License along with
    CASToR (in file GNU_GPL.TXT). If not, see <http://www.gnu.org/licenses/>.

Copyright 2017-2024 all CASToR contributors listed below:

    --> Didier BENOIT, Claude COMTAT, Marina FILIPOVIC, Thibaut MERLIN, Mael MILLARDET, Simon STUTE, Valentin VIELZEUF, Zacharias CHALAMPALAKIS, Alexandre MERASLI, Tie LIU, Maxime TOUSSAINT, Aurélien COUSSAT, Damian TRYBEK

This is CASToR version 3.2.
*/

/*!
  \file
  \ingroup  optimizer
  \brief    Declaration of class iOptimizerADMMLim
*/

#ifndef iOptimizerADMMLim_HH
#define iOptimizerADMMLim_HH 1

#include "gVariables.hh"
#include "sAddonManager.hh"
#include "vOptimizer.hh"
#include "iPenaltyMarkovRandomField.hh"

#define ADMMLIM_NOT_ADAPTIVE 0
#define ADMMLIM_ADAPTIVE_ALPHA 1
#define ADMMLIM_ADAPTIVE_ALPHA_AND_TAU 2
#define ADMMLIM_NOT_DEFINED -1

#define ADMMLIM_DO_NOT_SAVE_SINOGRAMS 0
#define ADMMLIM_SAVE_SINOGRAMS 1

/*!
  \class   iOptimizerADMMLim
  \brief   This class implements the ADMM with non-negativity on projection space
  \details This class inherits from vOptimizer and implements the ADMM algorithm
           proposed by Lim et al. to ensure positivity of the projections.
*/
class iOptimizerADMMLim : public vOptimizer
{
  // -------------------------------------------------------------------
  // Constructor & Destructor
  public:
    /*!
      \fn      public iOptimizerADMMLim::iOptimizerADMMLim()
      \brief   The constructor of iOptimizerADMMLim
      \details This is the default and unique constructor. It does not take any parameter and
               its role is only to affect default values to each member of the class.
    */
    iOptimizerADMMLim();
    /*!
      \fn      public iOptimizerADMMLim::~iOptimizerADMMLim()
      \brief   The destructor of iOptimizerADMMLim
      \details This is the default and unique destructor. It does not take any parameter and
               its role is only to free or delete all structures that were built by this class.
    */
    ~iOptimizerADMMLim();


  // -------------------------------------------------------------------
  // Public member functions
  public:
    // Function for automatic insertion (put the class name as the parameter and do not add semi-column at the end of the line)
    FUNCTION_OPTIMIZER(iOptimizerADMMLim)
    /*!
      \fn      public int iOptimizerADMMLim::ReadConfigurationFile()
      \param   const string& a_configurationFile
      \brief   A function used to read options from a configuration file
      \details This function implements the reading of all options associated to the child optimizer, from
               a configuration file. It is the implementation of the pure virtual function inherited
               from the abstract class vOptimizer. It checks the reading status but not
               the options values that will be checked by the CheckSpecificParameters() function.
      \return  An integer reflecting the reading success; 0 if success, another value otherwise.
    */
    int ReadConfigurationFile(const string& a_configurationFile);
    /*!
      \fn      public int iOptimizerADMMLim::ReadOptionsList()
      \param   const string& a_optionsList
      \brief   A function used to read options from a list of options
      \details This function implements the reading of all options associated to the child optimizer, from
               a list of options. It is the implementation of the pure virtual function inherited
               from the abstract class vOptimizer. It checks the reading status but not
               the options values that will be checked by the CheckSpecificParameters() function.
      \return  An integer reflecting the reading success; 0 if success, another value otherwise.
    */
    int ReadOptionsList(const string& a_optionsList);
    
    /*!
      \fn      public virtual int vOptimizer::DataStep4Optional()
      \param   oProjectionLine* ap_Line
      \param   vEvent* ap_Event
      \param   int a_bed
      \param   int a_timeFrame
      \param   int a_respGate
      \param   int a_cardGate
      \param   int a_thread
      \brief   A public function used to initialize u and v variables, and to compute projection of the gradient. 
      \details Inside the DataUpdateStep() of the oOptimizerManager, this is the second function called. This function is 
               overloaded ADMMLim optimizer to initialize u and v variables, and computes projection of the gradient. 
      \return  An integer reflecting the process status; 0 if no problem, another value otherwise.
    */
    virtual int DataStep4Optional( oProjectionLine* ap_Line, vEvent* ap_Event,
                                   int a_bed, int a_timeFrame, int a_respGate, int a_cardGate,
                                   int a_thread );
    /*!
      \fn      public virtual int iOptimizerADMMLim::DataStep5ComputeCorrections()
      \param   oProjectionLine* ap_Line
      \param   vEvent* ap_Event
      \param   int a_bed
      \param   int a_timeFrame
      \param   int a_respGate
      \param   int a_cardGate
      \param   int a_thread
      \brief   A public function used to compute the correction terms in the data space, for the provided event
      \details Inside the DataUpdateStep() of the oOptimizerManager, this is the fifth function called. Its role is to compute
               the correction terms in the data space, based on the forward model and the data. Here, it is overrided
               to get Ax, y, r_bar for u and v computation

      \return  An integer reflecting the process status; 0 if no problem, another value otherwise.
    */
    virtual int DataStep5ComputeCorrections( oProjectionLine* ap_Line, vEvent* ap_Event,
                                             int a_bed, int a_timeFrame, int a_respGate, int a_cardGate,
                                             int a_thread );
    /*!
      \fn      public int iOptimizerADMMLim::DataStep6Optional()
      \param   oProjectionLine* ap_Line
      \param   vEvent* ap_Event
      \param   int a_bed
      \param   int a_timeFrame
      \param   int a_respGate
      \param   int a_cardGate
      \param   int a_thread
      \brief   A public function to compute the analytical solution of ADMM equation on v and update u
      \details Inside the DataUpdateStep() of the oOptimizerManager, this is the sixth function called. 
               For this optimizer, here is computed u^{k+1} and v^{k+1} update thanks to the paper analytical formulas,
               using preivously stored variables and x update. 
      \return  An integer reflecting the process status; 0 if no problem, another value otherwise.
    */
    virtual int DataStep6Optional( oProjectionLine* ap_Line, vEvent* ap_Event,
                                   int a_bed, int a_timeFrame, int a_respGate, int a_cardGate,
                                   int a_thread );
  
  // -------------------------------------------------------------------
  // Private member functions (virtual in vOptimizer)
  private:
  /*!
      \fn      private virtual int iOptimizerADMMLim::PreDataUpdateSpecificStep()
      \brief   A private function used to compute some norms for the image update.
      \details This function computes some norms for the optimized stepsize used for the image update.
               so it can be overloaded by the child optimizer if needed.
      \return  An integer reflecting the process status; 0 if no problem, another value otherwise.
    */
    virtual int PreDataUpdateSpecificStep();
    /*!
      \fn      private int iOptimizerADMMLim::PreImageUpdateSpecificStep()
      \brief   A private function used to compute the penalty term update adaptive alpha of the ADMM algorithm
      \details This function implements the virtual eponym function of vOptimizer.
               It computes the penalty term of the ADMM algorithm, and the residuals to compute adaptive alpha.
      \return  An integer reflecting the process status; 0 if no problem, another value otherwise.
    */
    int PreImageUpdateSpecificStep();
    /*!
      \fn      private void iOptimizerADMMLim::ShowHelpSpecific()
      \brief   A function used to show help about the child optimizer
      \details This function must describe what the module does and how to use it. It describes in
               details the different parameters of the optimizer, and how to set them through the use
               of a configuration file or a list of options. It is the implementation of the pure
               virtual function inherited from the abstract class vOptimizer. It is called by the
               public ShowHelp() function.
    */
    void ShowHelpSpecific();
    /*!
      \fn      private int iOptimizerADMMLim::CheckSpecificParameters()
      \brief   A private function used to check the parameters settings specific to the child optimizer
      \details This function is used to check that all parameters specific to the optimizer are correctly set
               within allowed values. It is called by the CheckParameters() function of the mother class.
               It is the implementation of the pure virtual function inherited from the abstract mother
               class vOptimizer.
      \return  An integer reflecting the check status; 0 if no problem, another value otherwise.
    */
    int CheckSpecificParameters();
    /*!
      \fn      private int iOptimizerADMMLim::InitializeSpecific()
      \brief   This function is used to initialize specific stuff to the child optimizer.
      \details It is called by the public Initialize() function from the mother.
      \return  An integer reflecting the initialization status; 0 if no problem, another value otherwise.
    */
    int InitializeSpecific();
    /*!
      \fn      private int iOptimizerADMMLim::SensitivitySpecificOperations()
      \param   FLTNB a_data
      \param   FLTNB a_forwardModel
      \param   FLTNB* ap_weight
      \param   FLTNB a_multiplicativeCorrections
      \param   FLTNB a_additiveCorrections
      \param   FLTNB a_blankValue
      \param   FLTNB a_quantificationFactor
      \param   oProjectionLine* ap_Line
      \brief   This function compute the weight associated to the provided event (for sensitivity computation)
      \details It is the implementation of the pure virtual function from vOptimizer. The result is
               put at ap_weight location.
      \return  An integer reflecting the process status; 0 if no problem, another value otherwise.
    */
    int SensitivitySpecificOperations( FLTNB a_data, FLTNB a_forwardModel, FLTNB* ap_weight,
                                       FLTNB a_multiplicativeCorrections, FLTNB a_additiveCorrections, FLTNB a_blankValue,
                                       FLTNB a_quantificationFactor, oProjectionLine* ap_Line );
    /*!
      \fn      private int iOptimizerADMMLim::DataSpaceSpecificOperations()
      \param   FLTNB a_data
      \param   FLTNB a_forwardModel
      \param   FLTNB* ap_backwardValues
      \param   FLTNB a_multiplicativeCorrections
      \param   FLTNB a_additiveCorrections
      \param   FLTNB a_blankValue
      \param   FLTNB a_quantificationFactor
      \param   oProjectionLine* ap_Line
      \param   int a_th
      \brief   This function performs the data space operations specific to the optimizer (computes the values
               to be backprojected)
      \details It is the implementation of the pure virtual function from vOptimizer. It is not useful in this optimizer
               this function is directly written in DataStep5ComputeCorrections function as used variables depend on the current thread
      \return  An integer reflecting the process status; 0 if no problem, another value otherwise.
    */
    int DataSpaceSpecificOperations( FLTNB a_data, FLTNB a_forwardModel, FLTNB* ap_backwardValues,
                                     FLTNB a_multiplicativeCorrections, FLTNB a_additiveCorrections, FLTNB a_blankValue,
                                     FLTNB a_quantificationFactor, oProjectionLine* ap_Line);
    /*!
      \fn      private int iOptimizerADMMLim::ImageSpaceSpecificOperations()
      \param   FLTNB a_currentImageValue
      \param   FLTNB* apImageValue
      \param   FLTNB a_sensitivity
      \param   FLTNB* ap_correctionValues
      \param   INTNB a_voxel
      \param   int a_tbf
      \param   int a_rbf
      \param   int a_cbf
      \param   int a_th
      \brief   This function perform the image update step specific to the optimizer
      \details It is the implementation of the pure virtual function from vOptimizer. The new image value is
               put at the apImageValue location.
      \return  An integer reflecting the process status; 0 if no problem, another value otherwise.
    */
    int ImageSpaceSpecificOperations( FLTNB a_currentImageValue, FLTNB* apImageValue,
                                      FLTNB a_sensitivity, FLTNB* ap_correctionValues,
                                      INTNB a_voxel, int a_tbf = -1, int a_rbf = -1, int a_cbf = -1 );
    /*!
      \fn      public virtual int vOptimizer::ImageUpdateStep()
      \param   int a_iteration
      \param   int a_nbSubsets
      \brief   A public function used to perform the image update step of the optimizer
      \details This function is called by the eponym function from the oOptimizerManager. It will manage the dynamic loops,
               compute the sensitivity using the private ComputeSensitivity() function, and update each voxel according to
               the specific optimizer by calling the pure virtual ImageSpaceSpecificOperations() function.
      \return  An integer reflecting the process status; 0 if no problem, another value otherwise.
    */
    virtual int ImageUpdateStep();

  // -------------------------------------------------------------------
  // Data members
  private:
    FLTNB**** m4p_firstDerivativePenaltyImage;     /*!< Image containing the penalty terms of the algorithm */
    // Variables to compute u and v in DataStep6
    FLTNB** m2p_rBackgroundEvents;                 /*!< r_bar variable in original paper, to store background events (scatter + randoms), indexed on current thread and TOF bin */
    FLTNB** m2p_yData;                             /*!< y variable in original paper, to store raw data, indexed on current thread and TOF bin */
    // Variables involved in adaptive alpha and tau computation
    FLTNB m_alpha;                                 /*!< Alpha variable in original paper (ADMM penalty parameter), computed at each iteration if adaptive mode */
    FLTNB m_previousAlpha;                         /*!< Alpha variable at previous iteration, used to scale u if adaptive mode */
    FLTNB m_tau;                                   /*!< Tau variable in Wohlberg's paper, computed at each iteration if adaptive mode for alpha and tau */
    FLTNB m_mu;                                    /*!< Mu variable in adaptive alpha computation, factor to balance primal and dual residual in adaptive alpha computation */
    FLTNB m_tauMax;                                /*!< Tau_max variable in adaptive alpha and tau computation */
    FLTNB m_xi;                                    /*!< Xi variable in adaptive alpha computation, factor to balance primal and dual residual convergence speed in adaptive computations */
    // All stored vectors (those linked to gradient, sinograms u, v and Ax)
    FLTNB**** m4p_currentGrad;                     /*!< Current gradient vector (indexed on basis functions) to compute gradient norm in gradient descent in next x update and needed to compute projection of it */
    FLTNB**** m4p_currentGradFrameGates;           /*!< Current gradient vector (indexed on frames) */
    FLTNB** m2p_toWrite_uk;                        /*!< Sinogram u (scaled dual variable), to be used for next iteration and to be written at the end of optimizer computation, with u the scaled dual variable in second ADMM derivation */    
    FLTNB** m2p_toWrite_vk;                        /*!< Sinogram v (ADMM constraint is Ax=v), to be used for next iteration and to be written at the end of optimizer computation, with v=Ax in second ADMM derivation */
    FLTNB** m2p_vectorAx;                          /*!< Sinogram containing the forward projection of the image (for each lor and each TOF bin) */
    // Vector norms
    HPFLTNB* mp_gradSquareNorm;                    /*!< Norm of current gradient indexed on basis functions, to be used in gradient descent (to compute stepsize) in x update */
    HPFLTNB* mp_projGradSquareNorm;                /*!< Norm of projection of current gradient, to be used in gradient descent (to compute stepsize) in x update */
    HPFLTNB* mp_penaltyGradSquareNorm;             /*!< Norm of finite difference matrix times gradient vector, to be used in gradient descent with quadratic MRF penalty in x update */
    FLTNB m_relativePrimalSquareNorm;              /*!< Norm of relative primal residual (used in adaptive modes) */
    FLTNB m_relativeDualSquareNorm;                /*!< Norm of relative dual residual (used in adaptive modes) */
    HPFLTNB* mp_primalSquareNorm;                  /*!< Norm of primal residual (used in adaptive modes) */
    HPFLTNB* mp_squareNorm_Ax;                     /*!< Norm of projection of image (used in adaptive modes) */
    HPFLTNB* mp_squareNorm_v;                      /*!< Norm of sinogram v (used in adaptive modes) */
    // General variables
    int m_adaptiveParameters;                      /*!< Integer to define which variables are adaptive : nothing, alpha, or tau (which means alpha and tau) */
    FLTNB m_stoppingCriterionValue;                /*!< Value to use as stopping criterion based on relative residuals (0 means do not used the stopping criterion)*/
    int m_saveSinogramsUAndV;                      /*!< Integer to save (or not) sinograms u and v at same iterations as the image */
    int m_criterionReachedIteration;               /*!< Integer to define at which iteration the stopping criterion is reached, to stop the algorithm (if asked) */
    bool m_criterionReachedIsPassed;               /*!< Boolean to define when stopping criterion is reached, to stop the algorithm (if asked) */
    int m_firstIteration;                          /*!< Integer to define value of first computed iteration (if some are skipped), for u and v initialization */
    bool m_firstIterationIsPassed;                 /*!< Boolean to check if value of first computed iteration is set */
    bool u_and_v_from_user;                        /*!< Boolean to check if user gave sinograms u and v in command line */
};


// Class for automatic insertion (set here the visible optimizer's name as the first parameter,
// put the class name as the second parameter and do NOT add semi-colon at the end of the line)
CLASS_OPTIMIZER(ADMMLim,iOptimizerADMMLim)

#endif

