/*
This file is part of CASToR.

    CASToR is free software: you can redistribute it and/or modify it under the
    terms of the GNU General Public License as published by the Free Software
    Foundation, either version 3 of the License, or (at your option) any later
    version.

    CASToR is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
    details.

    You should have received a copy of the GNU General Public License along with
    CASToR (in file GNU_GPL.TXT). If not, see <http://www.gnu.org/licenses/>.

Copyright 2017-2024 all CASToR contributors listed below:

    --> Didier BENOIT, Claude COMTAT, Marina FILIPOVIC, Thibaut MERLIN, Mael MILLARDET, Simon STUTE, Valentin VIELZEUF, Zacharias CHALAMPALAKIS, Alexandre MERASLI, Tie LIU, Maxime TOUSSAINT, Aurélien COUSSAT, Damian TRYBEK

This is CASToR version 3.2.
*/

/*!
  \file
  \ingroup datafile

  \brief Implementation of class vEvent
*/

#include "vEvent.hh"
#include "vDataFile.hh"
#include "sOutputManager.hh"

// =====================================================================
// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// =====================================================================

vEvent::vEvent()
{
  m_timeInMs = 0;
  m_nbLines = 0;
  mp_ID1 = NULL;
  mp_ID2 = NULL;
  m_dataType = TYPE_UNKNOWN;
  m_dataMode = MODE_UNKNOWN;
  m_verbose = -1;
  m_eventValue = 0.;
  m_eventIndex = -1;
  m_nbCustomINTData = -1;
  m_nbCustomFLTData = -1;
  mp_customINTData = NULL;
  mp_customFLTData = NULL;
}

// =====================================================================
// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// =====================================================================

vEvent::~vEvent() 
{
  if (mp_ID1 != NULL) free(mp_ID1);
  if (mp_ID1 != NULL) free(mp_ID2);
  if (mp_customINTData != NULL) free(mp_customINTData);
  if (mp_customFLTData != NULL) free(mp_customFLTData);
}

// =====================================================================
// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// =====================================================================

int vEvent::AllocateID() 
{
  // Verbose
  DEBUG_VERBOSE(m_verbose,VERBOSE_DEBUG_LIGHT)
  if (m_verbose>=VERBOSE_DETAIL) Cout("vEvent::AllocateID() -> Allocate buffers for indices" << endl);
  // Check number of lines
  if (m_nbLines<1)
  {
    Cerr("***** vEvent::AllocateID() -> Error, number of lines has not been initialized (<1) !" << endl);
    return 1;
  }
  else
  {
    // Allocate buffer indices
    mp_ID1 = (uint32_t*)malloc(m_nbLines*sizeof(uint32_t));
    mp_ID2 = (uint32_t*)malloc(m_nbLines*sizeof(uint32_t));
  }
  // Call the pure virtual function implemented in child classes for the allocation of data specific to child classes
  if (AllocateSpecificData())
  {
    Cerr("***** vEvent::AllocateID() -> Error when trying to allocated specific data for the Event !" << endl);
    return 1;
  }
  // End
  return 0;
}

// =====================================================================
// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// =====================================================================
int vEvent::AllocateCustomINTData() 
{
  if(m_nbCustomINTData < 1)
  {
    Cerr("***** vEvent::AllocateCustomINTData() -> Error, number of custom EVTINTDATA data has not been initialized (<1) !" << endl);
    return 1;
  }
  else
    mp_customINTData = (EVTINTDATA*)malloc(m_nbCustomINTData*sizeof(EVTINTDATA));
  
  return 0;
}

// =====================================================================
// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// =====================================================================
int vEvent::AllocateCustomFLTData() 
{
  if(m_nbCustomFLTData < 1)
  {
    Cerr("***** vEvent::AllocateCustomFLTData() -> Error, number of custom EVTFLTDATA data has not been initialized (<1) !" << endl);
    return 1;
  }
  else
    mp_customFLTData = (EVTFLTDATA*)malloc(m_nbCustomFLTData*sizeof(EVTFLTDATA));
  
  return 0;
}

// =====================================================================
// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// =====================================================================
EVTFLTDATA vEvent::GetCustomFLTData(int a_idx)
{
  if(a_idx >= m_nbCustomFLTData || mp_customFLTData==NULL)
  {
    Cerr("***** vEvent::GetCustomFLTdata() -> Error when trying to recover custom EVTFLTDATA data #"<< a_idx+1 << "!" << endl);
    
    if(mp_customFLTData==NULL)
      Cerr("*****                              'mp_customFLTData' has not been initialized!" << endl);
    else 
      Cerr("*****                              the number of custom data is initialized to " << m_nbCustomFLTData << endl);
  }
  
  return mp_customFLTData[ a_idx ];
}


// =====================================================================
// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// =====================================================================
EVTINTDATA vEvent::GetCustomINTData(int a_idx)
{
  if(a_idx >= m_nbCustomINTData || mp_customINTData==NULL)
  {
    Cerr("***** vEvent::GetCustomINTdata() -> Error when trying to recover custom EVTINTDATA data #"<< a_idx+1 << "!" << endl);
    
    if(mp_customINTData==NULL)
      Cerr("*****                              'mp_customINTData' has not been initialized!" << endl);
    else 
      Cerr("*****                              the number of custom data is initialized to " << m_nbCustomINTData << endl);
  }
  
  return mp_customINTData[ a_idx ];
}

// =====================================================================
// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// =====================================================================
int vEvent::SetCustomFLTData(int a_idx, EVTFLTDATA a_value)
{
  if(a_idx >= m_nbCustomFLTData || mp_customFLTData==NULL)
  {
    Cerr("***** vEvent::SetCustomFLTData() -> Error when trying to set custom EVTFLTDATA data #"<< a_idx+1 << "!" << endl);
    
    if(mp_customFLTData==NULL)
      Cerr("*****                              'mp_customFLTData' has not been initialized!" << endl);
    else 
      Cerr("*****                              the number of custom data is initialized to " << m_nbCustomFLTData << endl);
      
    return 1;
  }
  
  mp_customFLTData[ a_idx ] = a_value;
  return 0;
}

// =====================================================================
// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// =====================================================================
int vEvent::SetCustomINTData(int a_idx, EVTINTDATA a_value)
{
  if(a_idx >= m_nbCustomINTData || mp_customINTData==NULL)
  {
    Cerr("***** vEvent::SetCustomINTData() -> Error when trying to set custom EVTINTDATA data #"<< a_idx+1 << "!" << endl);
    
    if(mp_customINTData==NULL)
      Cerr("*****                              'mp_customINTData' has not been initialized!" << endl);
    else 
      Cerr("*****                              the number of custom data is initialized to " << m_nbCustomINTData << endl);
      
    return 1;
  }
  
  mp_customINTData[ a_idx ] = a_value;
  return 0;
}
               
               
