/*
This file is part of CASToR.

    CASToR is free software: you can redistribute it and/or modify it under the
    terms of the GNU General Public License as published by the Free Software
    Foundation, either version 3 of the License, or (at your option) any later
    version.

    CASToR is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
    details.

    You should have received a copy of the GNU General Public License along with
    CASToR (in file GNU_GPL.TXT). If not, see <http://www.gnu.org/licenses/>.

Copyright 2017-2024 all CASToR contributors listed below:

    --> Didier BENOIT, Claude COMTAT, Marina FILIPOVIC, Thibaut MERLIN, Mael MILLARDET, Simon STUTE, Valentin VIELZEUF, Zacharias CHALAMPALAKIS, Alexandre MERASLI, Tie LIU, Maxime TOUSSAINT, Aurélien COUSSAT, Damian TRYBEK

This is CASToR version 3.2.
*/

/*!
  \file
  \ingroup utils
  \brief  This program is used to merge datafiles.
*/

#include "gVariables.hh"
#include "gOptions.hh"
#include "iEventHistoPET.hh"
#include "iEventNorm.hh"
#include "iEventPET.hh"
#include "iProjectorClassicSiddon.hh"
#include "oImageDimensionsAndQuantification.hh"
#include "oInterfileIO.hh"
#include "oProjectionLine.hh"
#include "oProjectorManager.hh"
#include "vDataFile.hh"
#include "iDataFilePET.hh"
#include "iDataFileSPECT.hh"
#include "iDataFileCT.hh"
#include "sOutputManager.hh"
#include "sRandomNumberGenerator.hh"
#include <algorithm>
#include <cstdint>
#include <omp.h>

// =============================================================================================================================================
// =============================================================================================================================================
// =============================================================================================================================================
//                                                        H E L P     F U N C T I O N S
// =============================================================================================================================================
// =============================================================================================================================================
// =============================================================================================================================================


/*!
  \fn      ShowHelp()
  \brief   Display main command line options for castor-datafileMerger
*/
void ShowHelp()
{
  // Show help
  cout << endl;
  cout << "Usage:  castor-includeNormIntoDatafilePET -df datafile.cdh -norm path_to_norm.hdr -(f/d)out output" << endl;
  cout << endl;
  cout << "This program can be used to include normalization factors from a normalization datafile into a PET datafile." << endl;
  cout << endl;
  cout << "[Mandatory parameters]:" << endl;
  cout << "  -df datafilePET.cdh    : Give a CASToR PET datafile" << endl;
  cout << "  -norm datafileNorm.cdh : Give a CASToR normalization datafile (the scanner should correspond to the one provided with -df)" << endl;
  cout << "  -fout name             : Give the root name for all output files (no default, alternative to -dout)" << endl;
  cout << "  -dout name             : Give the name of the output directory where all output files will be written (no default, alternative to -fout)" << endl;
  cout << endl;
  cout << "[Options]:" << endl;
  cout << "  -vb value              : Give the verbosity level, from 0 (no verbose) to 2 (default: 1)" << endl;
  cout << "  --help,-h,-help        : Print out this help page." << endl; // managed by main
  cout << endl;
  #ifdef BUILD_DATE
  cout << "  Build date: " << BUILD_DATE << endl;
  cout << endl;
  #endif
  #ifdef CASTOR_VERSION
  cout << "  This program is part of the CASToR release version " << CASTOR_VERSION << "." << endl;
  cout << endl;
  #endif
}

// =============================================================================================================================================
// =============================================================================================================================================
// =============================================================================================================================================
//                                                        M A I N     P R O G R A M
// =============================================================================================================================================
// =============================================================================================================================================
// =============================================================================================================================================

int main(int argc, char** argv)
{
  // ============================================================================================================
  // MPI stuff (we make all instances except the first one returning 0 directly)
  // ============================================================================================================
  int mpi_rank = 0;
  #ifdef CASTOR_MPI
  int mpi_size = 1;
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
  if (mpi_rank!=0) return 0;
  #endif

  // No argument, then show help
  if (argc==1)
  {
    ShowHelp();
    Exit(EXIT_SUCCESS);
  }

  // ============================================================================================================
  // Parameterized variables with their default values
  // ============================================================================================================

  // --------------------------------------------------------------------------------
  // Input settings
  // --------------------------------------------------------------------------------

  // Path to the normalization datafile
  string path_to_datafile = "";

  // Path to the normalization datafile
  string path_to_normalization_datafile = "";

  // --------------------------------------------------------------------------------
  // Output settings
  // --------------------------------------------------------------------------------

  // Output directory name.
  string path_dout = "";
  // Or root name
  string path_fout = "";

  // --------------------------------------------------------------------------------
  // Miscellaneous settings
  // --------------------------------------------------------------------------------

  // Verbose level
  int verbose = 1;

  // ============================================================================================================
  // Read command-line parameters
  // ============================================================================================================

  // Must manually increment the option index when an argument is needed after an option
  for (int i=1; i<argc; i++)
  {
    // Get the option as a string
    string option = (string)argv[i];

    // --------------------------------------------------------------------------------
    // Miscellaneous settings
    // --------------------------------------------------------------------------------

    // Show help
    if (option=="-h" || option=="--help" || option=="-help")
    {
      ShowHelp();
      Exit(EXIT_SUCCESS);
    }
    // General verbosity level
    else if (option=="-vb")
    {
      if (i>=argc-1)
      {
        Cerr("***** castor-datafileMerger() -> Argument missing for option: " << option << endl);
        Exit(EXIT_FAILURE);
      }
      if (ConvertFromString(argv[i+1], &verbose))
      {
        Cerr("***** castor-datafileMerger() -> Exception when trying to read provided verbosity level '" << verbose << " for option: " << option << endl);
        Exit(EXIT_FAILURE);
      }
      i++;
    }

    // --------------------------------------------------------------------------------
    // Input settings
    // --------------------------------------------------------------------------------

    // Datafile
    else if (option=="-df") // This is a mandatory option
    {
      if (i>=argc-1)
      {
        Cerr("***** castor-datafileMerger() -> Argument missing for option: " << option << endl);
        Exit(EXIT_FAILURE);
      }
      path_to_datafile = (string)argv[i+1];
      i++;
    }
    // Normalization datafile
    else if (option=="-norm") // This is a mandatory option
    {
      if (i>=argc-1)
      {
        Cerr("***** castor-datafileMerger() -> Argument missing for option: " << option << endl);
        Exit(EXIT_FAILURE);
      }
      path_to_normalization_datafile = (string)argv[i+1];
      i++;
    }

    // --------------------------------------------------------------------------------
    // Output settings
    // --------------------------------------------------------------------------------

    // Name of the output directory
    else if (option=="-dout") // This is a mandatory option
    {
      if (i>=argc-1)
      {
        Cerr("***** castor-datafileMerger() -> Argument missing for option: " << option << endl);
        Exit(EXIT_FAILURE);
      }
      path_dout = argv[i+1];
      i++;
    }
    // Base name of the output files
    else if (option=="-fout") // This is a mandatory option
    {
      if (i>=argc-1)
      {
        Cerr("***** castor-datafileMerger() -> Argument missing for option: " << option << endl);
        Exit(EXIT_FAILURE);
      }
      path_fout = argv[i+1];
      i++;
    }

    // --------------------------------------------------------------------------------
    // Unknown option!
    // --------------------------------------------------------------------------------

    else
    {
      Cerr("***** castor-datafileMerger() -> Unknown option '" << option << "' !" << endl);
      Exit(EXIT_FAILURE);
    }
  }

  // ============================================================================================================
  // Some checks
  // ============================================================================================================

  // Output files
  if (path_fout.empty() && path_dout.empty())
  {
    Cerr("***** castor-datafileMerger() -> Please provide an output option for output files (-fout or -dout) !" << endl);
    Exit(EXIT_FAILURE);
  }
  // Check that only one option has been provided
  if (!path_fout.empty() && !path_dout.empty())
  {
    Cerr("***** castor-datafileMerger() -> Please provide either output option -fout or -dout but not both !" << endl);
    Exit(EXIT_FAILURE);
  }

  // ============================================================================================================
  // Create sOutputManager
  // ============================================================================================================

  sOutputManager* p_OutputManager = sOutputManager::GetInstance();  
  // Set verbose level
  p_OutputManager->SetVerbose(verbose);
  // Set MPI rank
  p_OutputManager->SetMPIRank(mpi_rank);
  // Initialize output directory and base name
  if (p_OutputManager->InitOutputDirectory(path_fout, path_dout))
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred while initializing output directory !" << endl);
    Exit(EXIT_FAILURE);
  }
  // Log command line
  if (p_OutputManager->LogCommandLine(argc,argv))
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred while logging command line arguments !" << endl);
    Exit(EXIT_FAILURE);
  }

  // ============================================================================================================
  // Create sScannerManager
  // ============================================================================================================

  // Get user endianness (interfile I/O)
  GetUserEndianness();

  // Create the scanner manager
  sScannerManager* p_ScannerManager = sScannerManager::GetInstance();  
  p_ScannerManager->SetVerbose(verbose);

  // Get system name from the dataFile
  string scanner_name = "";
  if (ReadDataASCIIFile(path_to_datafile, "Scanner name", &scanner_name, 1, KEYWORD_MANDATORY))
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred while trying to find the system name in the datafile header !" << endl);
    Exit(EXIT_FAILURE);
  } 
  if (p_ScannerManager->FindScannerSystem(scanner_name) )
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred while searching for scanner system !" << endl);
    Exit(EXIT_FAILURE);
  }
  if (p_ScannerManager->BuildScannerObject() )
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred during scanner object construction ! !" << endl);
    Exit(EXIT_FAILURE);
  }
  // This toolkit only works for PET datafiles
  if (p_ScannerManager->GetScannerType() != SCANNER_PET)
  {
    Cerr("***** castor-datafileMerger() -> Only PET scanners are currently supported !" << endl);
    Exit(EXIT_FAILURE);
  }

  // ============================================================================================================
  // oImageDimensionsAndQuantification settings
  // ============================================================================================================
  
  // Create a default image dimensions and quantification object
  oImageDimensionsAndQuantification* p_ID = new oImageDimensionsAndQuantification();
  p_ID->SetDefault();
  p_ID->SetVerbose(verbose);
  p_ID->SetNbBeds(1);
  // Threading
  int nb_threads = 1;
  #ifdef CASTOR_OMP
  nb_threads = omp_get_num_procs();
  if (verbose>=2) Cout("  --> Use " << nb_threads << " threads" << endl);
  #endif
  if(p_ID->SetNbThreads(to_string(nb_threads)))
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred while setting the number of threads !" << endl);
    Exit(EXIT_FAILURE);
  }
  if (p_ID->Initialize())
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred while initializing image dimensions !" << endl);
    Exit(EXIT_FAILURE);
  }

  // ============================================================================================================
  // Create datafile
  // ============================================================================================================

  // Create datafile
  iDataFilePET* p_Datafile = new iDataFilePET();
  // Set data file name
  p_Datafile->SetHeaderDataFileName(path_to_datafile);
  // Few parameters
  p_Datafile->SetImageDimensionsAndQuantification(p_ID);
  p_Datafile->SetBedIndex(0);
  p_Datafile->SetVerbose(verbose);
  // Read information from the header, without affecting the quantification (as no ImageDimensionsAndQuantification object is used here)
  bool do_not_affect_quantification = false;
  if (p_Datafile->ReadInfoInHeader(do_not_affect_quantification))
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred during datafile header reading ! Abort." << endl);
    Exit(EXIT_FAILURE);
  }
  if (p_Datafile->CheckParameters())
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred while checking datafile parameters ! Abort." << endl);
    Exit(EXIT_FAILURE);
  }
  if (p_Datafile->ComputeSizeEvent())
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred in datafile initialization ! Abort." << endl);
    Exit(EXIT_FAILURE);
  }
  if (p_Datafile->InitializeMappedFile())
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred in datafile initialization ! Abort." << endl);
    Exit(EXIT_FAILURE);
  }
  if (p_Datafile->PrepareDataFile())
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred in datafile preparation ! Abort." << endl);
    Exit(EXIT_FAILURE);
  }

  // ============================================================================================================
  // Create normalization datafile
  // ============================================================================================================

  // Normalization datafile
  iDataFilePET* p_NormDatafile = NULL;
  if (path_to_normalization_datafile != "")
  {
    p_NormDatafile = new iDataFilePET();
    // Set data file name
    p_NormDatafile->SetHeaderDataFileName(path_to_normalization_datafile);
    // Few parameters
    p_NormDatafile->SetImageDimensionsAndQuantification(p_ID);
    p_NormDatafile->SetBedIndex(0);
    p_NormDatafile->SetVerbose(verbose);
    // Read information from the header, without affecting the quantification (as no ImageDimensionsAndQuantification object is used here)
    do_not_affect_quantification = false;
    if (p_NormDatafile->ReadInfoInHeader(do_not_affect_quantification))
    {
      Cerr("***** castor-datafileMerger() -> A problem occurred during datafile header reading ! Abort." << endl);
      Exit(EXIT_FAILURE);
    }
    if (p_NormDatafile->CheckParameters())
    {
      Cerr("***** castor-datafileMerger() -> A problem occurred while checking datafile parameters ! Abort." << endl);
      Exit(EXIT_FAILURE);
    }
    if (p_NormDatafile->GetDataMode() != MODE_NORMALIZATION)
    {
      Cerr("***** castor-datafileMerger() -> Data mode of the datafile provided with -norm is not normalization ! Abort." << endl);
      Exit(EXIT_FAILURE);
    }
    if (p_NormDatafile->ComputeSizeEvent())
    {
      Cerr("***** castor-datafileMerger() -> A problem occurred in datafile initialization ! Abort." << endl);
      Exit(EXIT_FAILURE);
    }
    if (p_NormDatafile->InitializeMappedFile())
    {
      Cerr("***** castor-datafileMerger() -> A problem occurred in datafile initialization ! Abort." << endl);
      Exit(EXIT_FAILURE);
    }
    if (p_NormDatafile->PrepareDataFile())
    {
      Cerr("***** castor-datafileMerger() -> A problem occurred in datafile preparation ! Abort." << endl);
      Exit(EXIT_FAILURE);
    }
  }  

  // ============================================================================================================
  // Create the output datafile
  // ============================================================================================================

  // Create output datafile object
  iDataFilePET* p_OutputDatafile = new iDataFilePET();
  // Build output data file from the input one
  if (p_OutputDatafile->SetParametersFrom(p_Datafile))
  {
    Cerr("***** castor-datafileMerger() -> An error occurred while setting parameters of output file from input file !" << endl);
    Exit(EXIT_FAILURE);
  }
  p_OutputDatafile->SetImageDimensionsAndQuantification(p_ID);
  // Normalization
  if(p_NormDatafile)
  {
    p_OutputDatafile->SetNormCorrectionFlagOn();
  }
  // Init file
  if (p_OutputDatafile->PROJ_InitFile())
  {
    Cerr("***** castor-datafileMerger() -> An error occurred while opening file for writing !" << endl);
    Exit(EXIT_FAILURE);
  }
  p_OutputDatafile->SetNbEvents(p_Datafile->GetSize()); // set the correct number of events after PROJ_InitFile
  if (p_OutputDatafile->ComputeSizeEvent())
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred in datafile initialization ! Abort." << endl);
    Exit(EXIT_FAILURE);
  }
  if (p_OutputDatafile->PrepareDataFile())
  {
    Cerr("***** castor-datafileMerger() -> A problem occurred in datafile preparation ! Abort." << endl);
    Exit(EXIT_FAILURE);
  }

  // ============================================================================================================
  // Merge the datafiles
  // ============================================================================================================

  // Index normalization factors for fast retrieval
  map<pair<uint32_t, uint32_t>, FLTNBDATA>* p_NormIndex = NULL;
  if(p_NormDatafile)
  {
    // Verbose
    if (verbose>=1) Cout("castor-datafileMerger() -> Index normalization file" << endl);

    p_NormIndex = new map<pair<uint32_t, uint32_t>, FLTNBDATA>();
    iEventNorm* p_EventNorm = NULL;
    int64_t printing_index = 0;

    // Compute start and stop indices
    int64_t index = 0;
    int64_t index_start = 0;
    int64_t index_stop  = 0;
    p_NormDatafile->GetEventIndexStartAndStop(&index_start, &index_stop);

    #pragma omp parallel for private(index) schedule(static), num_threads(nb_threads)
    for (index = index_start; index < index_stop; index++)
    {

      // The thread index
      int th = 0;
      #ifdef CASTOR_OMP
      th = omp_get_thread_num();
      #endif
      
      // Verbose
      if (th==0 && verbose>=2)
      {
        if (printing_index%((int)(index_stop/100/nb_threads))==0)
        {
          int percent = (int)( ((FLTNB)(index)) * 100. / (((FLTNB) index_stop/nb_threads) ));
          cout << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b      "
              << percent << " %                    " << flush;
        }
        printing_index++;
      }

      p_EventNorm = (iEventNorm*) p_NormDatafile->GetEvent(index, th);
      FLTNBDATA normalization_factor = p_EventNorm->GetNormalizationFactor();
      // To save memory, we index only normalization factors whose value differs from 1
      if(normalization_factor != 1.)
      {
        uint32_t id1 = p_EventNorm->GetID1(0);
        uint32_t id2 = p_EventNorm->GetID2(0);

        // Order the IDs for disambiguation and making loop-up easier
        uint32_t id_lesser = min(id1, id2);
        uint32_t id_greater = max(id1, id2);

        // Index the value
        #pragma omp critical
        {
          p_NormIndex->insert({{id_lesser, id_greater}, normalization_factor});
        }
      }
    }

    // Verbose
    if (verbose>=2)
    {
      cout << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
          << "  --> 100 %                        " << endl;
      Cout("  --> Index normalization factors" << endl);
    }
  }

  // Verbose
  if (verbose>=1) Cout("castor-datafileMerger() -> Merge events" << endl);

  // Variables used in loop
  int64_t printing_index = 0;

  // Compute start and stop indices
  int64_t index = 0;
  int64_t index_start = 0;
  int64_t index_stop  = 0;
  p_Datafile->GetEventIndexStartAndStop(&index_start, &index_stop);

  #pragma omp parallel for private(index) schedule(static), num_threads(nb_threads)
  for (index = index_start; index < index_stop; index++)
  {
    // The thread index
    int th = 0;
    #ifdef CASTOR_OMP
    th = omp_get_thread_num();
    #endif
    
    // Verbose
    if (th==0 && verbose>=2)
    {
      if (printing_index%((int)(index_stop/100/nb_threads))==0)
      {
        int percent = (int)( ((FLTNB)(index)) * 100. / (((FLTNB) index_stop/nb_threads) ));
        cout << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b      "
             << percent << " %                    " << flush;
      }
      printing_index++;
    }

    // Read the event
    iEventPET* p_EventPET = (iEventPET*) p_Datafile->GetEvent(index, th);
    
    // Map normalization factor is present
    if (p_NormIndex)
    {
      uint32_t id1 = p_EventPET->GetID1(0);
      uint32_t id2 = p_EventPET->GetID2(0);

      // Order the IDs for disambiguation and making loop-up easier
      uint32_t id_lesser = min(id1, id2);
      uint32_t id_greater = max(id1, id2);

      pair<uint32_t, uint32_t> key = {id_lesser, id_greater};

      FLTNBDATA normalization_factor = 1.; // Default value
      // If value is found in the index, add it to the event
      if (p_NormIndex->count(key))
      {
        normalization_factor = p_NormIndex->at(key);
      }
      p_EventPET->SetNormalizationFactor(normalization_factor);
    }

    p_OutputDatafile->WriteEvent(p_EventPET, th);
  }

  // Verbose
  if (verbose>=2)
  {
    cout << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
         << "  --> 100 %                        " << endl;
    Cout("  --> Merge datafiles" << endl);
  }

  // ============================================================================================================
  // Write final output
  // ============================================================================================================

  // Write final file
  if (p_OutputDatafile->PROJ_WriteData())
  {
    Cerr("***** castor-datafileMerger() -> An error occurred while writing the final file !" << endl);
    Exit(EXIT_FAILURE);
  }
  // Delete temporary files
  if (p_OutputDatafile->PROJ_DeleteTmpDataFile())
  {
    Cerr("***** castor-datafileMerger() -> An error occurred while deleting temporary files !" << endl);
    Exit(EXIT_FAILURE);
  }
  // Write header
  if (p_OutputDatafile->WriteHeader())
  {
    Cerr("***** castor-datafileMerger() -> An error occurred while writing output header file !" << endl);
    Exit(EXIT_FAILURE);
  }

  // ============================================================================================================
  // Exit
  // ============================================================================================================

  // Delete objects in the inverse order in which they were created
  if (verbose>=5) Cout("----- Deleting CASToR objects ... -----" << endl);
  delete p_NormIndex;
  delete p_OutputDatafile;
  if (p_NormDatafile) delete p_NormDatafile;
  delete p_Datafile;
  delete p_ID;

  if (verbose>=5) Cout("----- CASToR objects successfully deleted -----" << endl);

  // Ending
  #ifdef CASTOR_MPI
  MPI_Finalize();
  #endif
  return EXIT_SUCCESS;
}
